#!/usr/bin/python
#
#   PROGRAM:     Firebird PowerConsole
#   MODULE:      common.py
#   DESCRIPTION: Common Classes and functions
#
#  The contents of this file are subject to the Initial
#  Developer's Public License Version 1.0 (the "License");
#  you may not use this file except in compliance with the
#  License. You may obtain a copy of the License at
#  http://www.ibphoenix.com/main.nfs?a=ibphoenix&page=ibp_idpl.
#
#  Software distributed under the License is distributed AS IS,
#  WITHOUT WARRANTY OF ANY KIND, either express or implied.
#  See the License for the specific language governing rights
#  and limitations under the License.
#
#  The Original Code was created by Pavel Cisar
#  for the Firebird Open Source RDBMS project.
#  http://www.firebirdsql.org
#
#  Copyright (c) 2006 Pavel Cisar <pcisar@users.sourceforge.net>
#  and all contributors signed below.
#
#  All Rights Reserved.
#  Contributor(s): ______________________________________.

import pyparsing as pp
import types
from fbpw.ibptoolbox import unquote, columnize
import kinterbasdb as kdb
from fbpw.fbcore import ObjectVisitor, TableColumn

#---Constants---#000000#FFFFFF--------------------------------------------------

HELP_PREFIX = 'help_'

RENDERER_LINE        = 0
RENDERER_LIST        = 1
RENDERER_OBJECT      = 2
RENDERER_OBJECTLIST  = 3
RENDERER_RESULT      = 4

#---Exceptions---#000000#FFFFFF-------------------------------------------------

class PWException(Exception):
    """An exception generated directly by PowerConsole.

    All exceptions thrown by PowerConsole should be derived from this class."""

    def __init__(self, message):
        """Construct a new 'PWException'.

        'message' -- A string describing the cause of the message."""

        Exception.__init__(self, message)

class PWError(PWException):
    """General programming error generated by PowerConsole."""
    pass

#---Classes---#000000#FFFFFF----------------------------------------------------

class pwHelpProvider(object):
    """Base class for all PowerConsole built-in help topic providers"""

    def __init__(self, pythonhelp):
        self._help = pythonhelp
        self.topics = []
        for attr in dir(self):
            if attr.startswith(HELP_PREFIX):
                self.topics.append(attr[len(HELP_PREFIX):])

    def getTopics(self):
        return self.topics
    def canHandle(self,topic):
        return hasattr(self, HELP_PREFIX+topic)
    def getHelp(self, topic):
        if hasattr(self, HELP_PREFIX+topic):
            return getattr(self, HELP_PREFIX+topic)
        return None

class pwCommand(object):
    """Base class for all PowerConsole built-in commands.
    
    Attributes:
    tooltip     Tuple or callable returning a tuple for tooltip in 
                the format (name, argspec, tip). Could be defined in descendants.
    """
    
    def __init__(self, shell):
        self.shell = shell
        self.__name = None
        
        # Basic grammar building blocks
        self.integer      = pp.Word(pp.nums+"+-",pp.nums).setName('integer')
        self.integer.setParseAction(self._tokenToInt)
        self.ident        = pp.Word(pp.alphas,pp.alphanums+'_.').setName('ident')
        #ident.setResultsName('expr')
        self.blobid       = (pp.Word(pp.hexnums) + ':' + pp.Word(pp.hexnums)).setName('blobid')
        self.filename     = (pp.Word(pp.alphas + pp.alphas8bit + pp.nums + '_/\-:.') | \
                            pp.QuotedString("'")).setName('filename')
        self.dbident      = (pp.Word(pp.alphas,pp.alphanums+'_$').setParseAction(self._upper) | 
                            pp.QuotedString('"')).setName('dbident')
        self.expr         = pp.White().suppress() + pp.CharsNotIn('\n').setName('expr')
        self.arg          = pp.CharsNotIn('\n').setName('arg')

#---Private---#000000#FFFFFF----------------------------------------------------

#---Protected---#000000#FFFFFF--------------------------------------------------

    def _getGlobals(self):
        return self.shell.commander.locals

    def _tokenToInt(self, st, loc, toks):
        return int(toks[0])

    def _upper(s, l, tokens):
        return tokens[0].upper()

    def _fail(self, s, loc, expr, err):
    	raise pp.ParseFatalException(s,loc,err,expr)

    def _cmdParseAction(s, l, tokens):
        params = []
        for par,value in tokens.items():
            if par == 'expr':
                s = "%s=%s" % (par,value)
            else:
                if type(value) == pp.ParseResults:
                    value = ''.join(value)
                if isinstance(value,types.StringTypes):
                    single = value.find("'")
                    double = value.find('"')
                    if (single != -1) and (double != -1):
                        s = '%s="""%s"""' % (par,value)
                    elif (single != -1) and (double == -1):
                        s = '%s="%s"' % (par,value)
                    else:
                        s = "%s='%s'" % (par,value)
                elif type(value) == types.IntType:
                    s = "%s=%i" % (par,value)
                else:
                    s = "%s='%s'" % (par,value)
            params.append(s)
        cmdCall = '_fbpw_execute(%s)' % (','.join(params))
        return cmdCall

#---Public---#000000#FFFFFF-----------------------------------------------------

    def getName(self):
        def _getName(grammar):
            if grammar == None:
                return None
            if hasattr(grammar,'exprs'):
                l = grammar.exprs
            else:
                l = [grammar,]
            for g in l:
                if g.resultsName == 'cmd':
                    return g.name
                elif len(g.exprs) > 0:
                    name = _getName(g)
                    if name != None:
                        return name
            return None
        if self.__name == None:
            name = _getName(self.getGrammar())
            if name != None:
                self.__name = unquote(name)
            else:
                raise PWError("Command %s doesn't have a name." % self.__class__.__name__)
        return self.__name
    def getGrammar(self):
        return None
    def getSQL(self):
        return None
    def getCmdKeyword(self):
        return None
    def execute(self,**kwargs):
        print "Method %s.execute not implemented" % self.__class__.__name__

class pwRenderer(object):
    """Base class for all PowerConsole renderers.
    """
    def __init__(self,shell):
        self.shell = shell
    def initialize(self):
        pass
    def finalize(self):
        pass

class pwLineRenderer(pwRenderer):
    """Default Line Renderer.
    
    Prints lines of test to the stdout.
    """
    def write(self,line=''):
        print line,
    def writeLine(self,line=''):
        print line
    def writeLines(self,lines):
        for line in lines:
            self.writeLine(line)
    
class pwListRenderer(pwRenderer):
    """Default List Renderer.
    
    Prints list of items to the stdout in columnized format.
    """
    def writeList(self,items):
        for line in columnize(items):
            print line

class pwObjectRenderer(pwRenderer,ObjectVisitor):
    """Default Object Renderer.
    
    Prints object's fullDescription to the stdout.
    """
    def visitDefaultAction(self,obj):
        print obj.name
    def visitCollation(self, collation):
        print "Print details for collation "+collation.getQuotedName()
    def visitCharacterSet(self, charset):
        print "Print details for character set "+charset.getQuotedName()
    def visitTable(self, table):
        print "Print details for table "+table.getQuotedName()
    def visitDependency(self,dependency):
        if dependency.field_name:
            print dependency.dependent_name,'depends on field',dependency.field_name,'in',dependency.depended_on_name
        else:
            print dependency.dependent_name,'depends on ',dependency.depended_on_name
    def writeObject(self,obj):
        obj.acceptVisitor(self)

class pwObjectListRenderer(pwRenderer,ObjectVisitor):
    """Default Object List Renderer.
    
    Prints obj.fullDescription or str(obj) to the stdout for each object in the list.
    """
    def visitDefaultAction(self,obj):
        print obj.name
    def visitDomain(self, table):
        print "Print details for domain "+table.getQuotedName()
    def visitTable(self, table):
        print "Print details for table "+table.getQuotedName()
    def visitCollation(self, collation):
        print "Print details for collation "+collation.getQuotedName()
    def visitCharacterSet(self, charset):
        print "Print details for character set "+charset.getQuotedName()
    def visitDependency(self,dependency):
        dependent = dependency.getDependent()
        dependency = dependency.getDependency()
        if isinstance(dependent,TableColumn):
            print 'Colum',dependent.name,'in',dependent.table_name,
        else:
            print dependent.__class__.__name__,dependent.name,
        print 'depends on',
        if isinstance(dependency,TableColumn):
            print 'colum',dependency.name,'in',dependency.table_name
        else:
            print dependency.__class__.__name__.lower(),dependency.name
    def writeList(self,items):
        for obj in items:
            obj.acceptVisitor(self)

class pwResultsetRenderer(pwRenderer):
    """Default Resultset Renderer.
    
    Prints data from KInterbasDB cursor to the stdout.
    """
    def writeResult(self, cursor, sql):
        if not sql.upper().startswith('SELECT'):
#            self.shell.writeErr('Not a select')
            return
#        print "SQL command:%s\n" % sql
#        print
        rowDesc = []
        for fieldDesc in cursor.description:
            rowDesc.append((fieldDesc[kdb.DESCRIPTION_NAME],
                max((len(fieldDesc[kdb.DESCRIPTION_NAME]),fieldDesc[kdb.DESCRIPTION_DISPLAY_SIZE]))))
        for fldName, fldSize in rowDesc:
            print fldName.ljust(fldSize),' ',
        print
        for fldName, fldSize in rowDesc:
            print "-" * fldSize,' ',
        print
        # For each row, print the value of each field left-justified within
        # the maximum possible width of that field.
        fieldIndices = range(len(cursor.description))
        for row in cursor:
            for fieldIndex in fieldIndices:
                fieldValue = str(row[fieldIndex])
                print fieldValue.ljust(rowDesc[fieldIndex][1]), ' ',
            print
