#!/usr/bin/python
#
#   PROGRAM:     IBPhoenix Python Toolbox
#   MODULE:      ibptoolbox.py
#   DESCRIPTION: Various helper functions and classes
#
#  The contents of this file are subject to the Initial
#  Developer's Public License Version 1.0 (the "License");
#  you may not use this file except in compliance with the
#  License. You may obtain a copy of the License at
#  http://www.ibphoenix.com/main.nfs?a=ibphoenix&page=ibp_idpl.
#
#  Software distributed under the License is distributed AS IS,
#  WITHOUT WARRANTY OF ANY KIND, either express or implied.
#  See the License for the specific language governing rights
#  and limitations under the License.
#
#  The Original Code was created by Pavel Cisar
#  for the Firebird Open Source RDBMS project.
#  http://www.firebirdsql.org
#
#  Copyright (c) 2006 Pavel Cisar <pcisar@users.sourceforge.net>
#  and all contributors signed below.
#
#  All Rights Reserved.
#  Contributor(s): Tim Delaney.

import types
import sys

def MixIn(pyClass, mixInClass, makeAncestor=False):
   if makeAncestor:
     if mixInClass not in pyClass.__bases__:
        pyClass.__bases__ = (mixInClass,) + pyClass.__bases__
   else:
     # Recursively traverse the mix-in ancestor
     # classes in order to support inheritance
     baseClasses = list(mixInClass.__bases__)
     baseClasses.reverse()
     for baseClass in baseClasses:
        MixIn(pyClass, baseClass)
     # Install the mix-in methods into the class
     for name in dir(mixInClass):
        if not name.startswith('__'):
        # skip private members
           member = getattr(mixInClass, name)
           if type(member) is types.MethodType:
               member = member.im_func
           setattr(pyClass, name, member)

from glob import glob
import os
def MixInDir(dir):
    ''' Installs mix-ins found in the given directory.
          Each mix-in is expected to match the name
          of a class in our domain classes. '''
    path = os.path.join(basePath, '*.py')
    for filename in glob(path):
          dirname, name = os.path.split(filename)
          name = os.path.splitext(name)[0]
          # Get the real class
          module = __import__(name)
          pyClass = getattr(module, name)
          # Get the mix-in class
          results = {}
          exec open(filename) in results
          mixInClass = results[name]
          # Mix it in
          MixIn(pyClass, mixInClass)

def importModule(name):
    mod = __import__(name)
    components = name.split('.')
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return mod

def unquote(str):
    if (str[-1:] == '"') and (str[:1] == '"'):
        return str[1:-1]
    else:
        return str

def isiterable(obj):
    return hasattr(obj, '__iter__') and callable(obj.__iter__)

def iif(expr,valueA,valueB):
    if expr: return valueA
    else: return valueB

def update_meta (self, other):
    self.__name__ = other.__name__
    self.__doc__ = other.__doc__
    self.__dict__.update(other.__dict__)
    return self

def columnize(list, displaywidth=80):
    """Generator function that returns a list of strings as a compact set of columns.

    Each column is only as wide as necessary.
    Columns are separated by three spaces (one was not legible enough).
    """
    if not list:
        print "<empty>"
        return
    nonstrings = [i for i in range(len(list))
                    if not isinstance(list[i],types.StringTypes)]
    if nonstrings:
        raise TypeError, ("list[i] not a string for i in %s" %
                          ", ".join(map(str, nonstrings)))
    size = len(list)
    if size == 1:
        print '%s' % str(list[0])
        return
    # Try every row count from 1 upwards
    for nrows in range(1, len(list)):
        ncols = (size+nrows-1) // nrows
        colwidths = []
        totwidth = -3
        for col in range(ncols):
            colwidth = 0
            for row in range(nrows):
                i = row + nrows*col
                if i >= size:
                    break
                x = list[i]
                colwidth = max(colwidth, len(x))
            colwidths.append(colwidth)
            totwidth += colwidth + 3
            if totwidth > displaywidth:
                break
        if totwidth <= displaywidth:
            break
    else:
        nrows = len(list)
        ncols = 1
        colwidths = [0]
    lines = []
    for row in range(nrows):
        texts = []
        for col in range(ncols):
            i = row + nrows*col
            if i >= size:
                x = ""
            else:
                x = list[i]
            texts.append(x)
        while texts and not texts[-1]:
            del texts[-1]
        for col in range(len(texts)):
            texts[col] = texts[col].ljust(colwidths[col])
        yield "%s" % str("   ".join(texts))

class LateBindingProperty (property):
    """Recipe from Tim Delaney, 2005/03/31
       http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/408713
       
class C(object):
    
    def getx(self):
        print 'C.getx'
        return self._x

    def setx(self, x):
        print 'C.setx'
        self._x = x

    def delx(self):
        print 'C.delx'
        del self._x

    x = LateBindingProperty(getx, setx, delx)

class D(C):

    def setx(self, x):
        print 'D.setx'
        super(D, self).setx(x)

    def delx(self):
        print 'D.delx'
        super(D, self).delx()

c = C()
c.x = 1
c.x
c.x
del c.x

print

d = D()
d.x = 1
d.x
d.x
del d.x

This has the advantages that:

a. You get back an actual property object (with attendant memory savings, performance increases, etc);

b. It's the same syntax as using property(fget, fset, fdel, doc) except for the name;

c. It will fail earlier (when you define the class as opposed to when you use it).

d. It's shorter ;)

e. If you inspect the property you will get back functions with the correct __name__, __doc__, etc.

    """
    def __new__(self, fget=None, fset=None, fdel=None, doc=None):
        if fget is not None:
            def __get__(obj, objtype=None, name=fget.__name__):
                fget = getattr(obj, name)
                return fget()
            fget = update_meta(__get__, fget)
        if fset is not None:
            def __set__(obj, value, name=fset.__name__):
                fset = getattr(obj, name)
                return fset(value)
            fset = update_meta(__set__, fset)
        if fdel is not None:
            def __delete__(obj, name=fdel.__name__):
                fdel = getattr(obj, name)
                return fdel()
            fdel = update_meta(__delete__, fdel)
        return property(fget, fset, fdel, doc)

#
# Decorator for class initializers, uses __metaclass__ hook.
# Created by Michele Simionato
# see http://www.phyast.pitt.edu/~micheles/python/classinitializer.html
#
# Example:
# import datetime
# 
# @classinitializer
# def def_properties(cls, schema):
#     """
#     Add properties to cls, according to the schema, which is a list
#     of pairs (fieldname, typecast). A typecast is a
#     callable converting the field value into a Python type.
#     The initializer saves the attribute names in a list cls.fields
#     and the typecasts in a list cls.types. Instances of cls are expected
#     to have private attributes with names determined by the field names.
#     """
#     cls.fields = []
#     cls.types = []
#     for name, typecast in schema:
#         if hasattr(cls, name): # avoid accidental overriding
#             raise AttributeError('You are overriding %s!' % name)
#         def getter(self, name=name):
#             return getattr(self, '_' + name)
#         def setter(self, value, name=name, typecast=typecast):
#             setattr(self, '_' + name, typecast(value))
#         setattr(cls, name, property(getter, setter))
#         cls.fields.append(name)
#         cls.types.append(typecast)
#
# def date(isodate): # add error checking if you like
#     "Convert an ISO date into a datetime.date object"
#     return datetime.date(*map(int, isodate.split('-')))
#
# class Article(object):
#     # fields and types are dynamically set by the initializer
#     def_properties([('title', str), ('author', str), ('date', date)])
#     def __init__(self, values): # add error checking if you like
#         for field, cast, value in zip(self.fields, self.types, values):
#             setattr(self, '_' + field, cast(value))
#

def classinitializer(proc):
  # basic idea stolen from zope.interface.advice, which credits P.J. Eby
  def newproc(*args, **kw):
      frame = sys._getframe(1)
      if '__module__' in frame.f_locals and not \
         '__module__' in frame.f_code.co_varnames: # we are in a class
          if '__metaclass__' in frame.f_locals:
            raise SyntaxError("Don't use two class initializers or\n"
            "a class initializer together with a__metaclass__ hook")
          def makecls(name, bases, dic):
             try:
                cls = type(name, bases, dic)
             except TypeError, e:
                if "can't have only classic bases" in str(e):
                   cls = type(name, bases + (object,), dic)
                else: # other strange errors, such as __slots__ conflicts, etc
                   raise
             proc(cls, *args, **kw)
             return cls
          frame.f_locals["__metaclass__"] = makecls
      else:
          proc(*args, **kw)
  newproc.__name__ = proc.__name__
  newproc.__module__ = proc.__module__
  newproc.__doc__ = proc.__doc__
  newproc.__dict__ = proc.__dict__
  return newproc

