#!/usr/bin/python
#
#   PROGRAM:     Firebird PowerConsole
#   MODULE:      pwcore.py
#   DESCRIPTION: PowerConsole core classes
#
#  The contents of this file are subject to the Initial
#  Developer's Public License Version 1.0 (the "License");
#  you may not use this file except in compliance with the
#  License. You may obtain a copy of the License at
#  http://www.ibphoenix.com/main.nfs?a=ibphoenix&page=ibp_idpl.
#
#  Software distributed under the License is distributed AS IS,
#  WITHOUT WARRANTY OF ANY KIND, either express or implied.
#  See the License for the specific language governing rights
#  and limitations under the License.
#
#  The Original Code was created by Pavel Cisar
#  for the Firebird Open Source RDBMS project.
#  http://www.firebirdsql.org
#
#  Copyright (c) 2006 Pavel Cisar <pcisar@users.sourceforge.net>
#  and all contributors signed below.
#
#  All Rights Reserved.
#  Contributor(s): ______________________________________.

from fbpw.ibptoolbox import importModule, unquote
import pyparsing as pp
import sys
import types
import string
from fbpw.common import *

#---Constants---#000000#FFFFFF--------------------------------------------------

IDENTCHARS = string.ascii_letters + string.digits + '_'

SQL_PARAM_SEPARATOR   = '<<'
SQL_STORE_SEPARATOR   = '>>'
SQLCMD = 'sql'

#---Classes---#000000#FFFFFF----------------------------------------------------

class CommandManager(object):
    def __init__(self,shell,**kwargs):
        self.shell = shell
        # Commands
        self.commands = {}
        self.term = ";"
        mod = importModule('fbpw.stdcmd')
        # Find and instantiate commands
        for item in dir(mod):
            if (type(mod.__dict__[item]) == types.TypeType) and \
                issubclass(mod.__dict__[item],pwCommand) and \
                (mod.__dict__[item] != pwCommand):
                cmd = mod.__dict__[item](shell)
                #print cmd.__class__.__name__
                self.commands[cmd.getName()] = cmd
        # Build Grammars
        self.__buildGrammar()
        self.__buildSQLCheckGrammar()
        self.__buildCommandCheckGrammar()

#---Private---#000000#FFFFFF----------------------------------------------------

    def __buildGrammar(self):
        self.grammarFull = None
        for cmd in self.commands.values():
            cmdGrammar = cmd.getGrammar()
            if cmdGrammar == None: continue
            if self.grammarFull != None:
                self.grammarFull = self.grammarFull | cmdGrammar
            else:
                self.grammarFull = cmdGrammar
        self.grammarFull = self.grammarFull.setResultsName('command') + pp.LineEnd()
    def __buildSQLCheckGrammar(self):
        self.grammarSQL = None
        for cmd in self.commands.values():
            cmdGrammar = cmd.getSQL()
            if not cmdGrammar: continue
            if self.grammarSQL:
                self.grammarSQL = self.grammarSQL | cmdGrammar
            else:
                self.grammarSQL = cmdGrammar
        self.grammarSQL = self.grammarSQL.setResultsName('command')
    def __buildCommandCheckGrammar(self):
        self.grammarCommands = None
        for cmd in self.commands.values():
            cmdGrammar = cmd.getCmdKeyword()
            if not cmdGrammar: continue
            if self.grammarCommands:
                self.grammarCommands = self.grammarCommands | cmdGrammar
            else:
                self.grammarCommands = cmdGrammar
        self.grammarCommands = self.grammarCommands.setResultsName('command')
    def __parseline(self, line):
        """Parse the line into a command name and a string containing
        the arguments.  Returns a tuple containing (command, args, line).
        'command' and 'args' may be None if the line couldn't be parsed.
        """
        i, n = 0, len(line)
        while i < n and line[i] in string.whitespace: i += 1
        indent, line = line[:i], line[i:].strip()
        if not line:
            return None, None, line
        i, n = 0, len(line)
        while i < n and line[i] in IDENTCHARS: i += 1
        cmd = line[:i]
        if self.isSQL(line):
            return indent, SQLCMD, line
        else:
            return indent, cmd, line

    def __onecmd(self, indent, cmd, line):
        """Checks whether the line is a command, and converts it into executable function call.
        """
#        print "indent",indent
#        print "cmd:",cmd
#        print "line:",line
        if not line:
            return ''
        elif not cmd:
            return line
        else:
            if self.isCmd(line):
                call = self.parseCmd(line)
                if call != None:
                    return indent + call
            # It's not a valid command
            return indent + line

#---Protected---#000000#FFFFFF--------------------------------------------------

#---Public---#000000#FFFFFF-----------------------------------------------------

    def debugPrint(self,*args):
    	self.shell.redirectStdout(False)
    	for arg in args:
    	    print arg,
    	print
    	self.shell.redirectStdout(True)

    def setCommands(self):
        self.locals = self.shell.interp.locals
        import __builtin__
        __builtin__._fbpw_execute = self.execute
        __builtin__.cmd = self

    def parseCmd(self, cmdstr):
        return self.grammarFull.parseString(cmdstr).command

    def isCmd(self, cmdstr):
        try:
            ret = self.grammarCommands.parseString(cmdstr)
            return True
        except pp.ParseException:
            return False

    def isSQL(self, cmdstr):
        try:
            ret = self.grammarSQL.parseString(cmdstr)
            return True
        except pp.ParseException:
            return False

    def getToolTip(self, command='', *args, **kwds):
        """Returns the tooltip for the command. Tooltip is a tuple of three
        values:
        
        cmd     - Command.
        args    - Arguments to be placed into the command line.
        text    - Text to be displayed in tooltip window.
        """
        if self.commands.has_key(command):
            cmd = self.commands[command]
            try:
                tip = getattr(cmd, 'tooltip')
            except AttributeError:
                return ('','','')
            if callable(tip):
                return tip()
            else:
                return tip
        else:
            return (None,None,None)

    def getAutoCompleteList(self, command):
        """Returns autocomplete list for the command.
        """
        # Todo: Implement autocomplete
        return []

    def execute(self,cmd,**kwargs):
        """Executes the command.
        """
        try:
            self.commands[cmd].execute(**kwargs)
        except Exception, e:
            self.shell.writeErr("%s : %s" % (str(e.__class__), str(e)))
            return

    def compileCommand(self, command):
        """Checks whether the command contains PowerConsole or SQL command, 
        and converts them into executable function call. Command could 
        have muiltiple lines and could be incomplete.
        
        Returns three values:
        cmd     - Command to be passed to the intrepreter
        more    - True when cmd is not a complete command (request additional line).
        stop    - True if there was an error in compilation, stop processing req.
        """
        self.debugPrint("Cmp:",command)
        lines = command.split('\n')
        result = []
        multilinesql = False
        i, n = 0, len(lines)
        try:
            while i < n:
                line = lines[i]
                if not multilinesql:
                    indent, cmd, line = self.__parseline(line)
#                    print "indent:",indent
#                    print "cmd:'",cmd,"'"
#                    print "line:'",line,"'"
                    if (cmd == SQLCMD):
                        sqlcmd = []
                        #sqlindent = indent
                        multilinesql = (line[-len(self.term):] != self.term)
                        if not multilinesql:
                            line = line[:-len(self.term)].strip()
                            result.append(self.__onecmd(indent, cmd, indent+SQLCMD+' '+line))
                        else:
                            sqlcmd.append(indent+SQLCMD+' '+line)
                    else:
                        result.append(self.__onecmd(indent, cmd, line))
                else:
                    multilinesql = (line[-len(self.term):] != self.term)
                    if not multilinesql:
                        line = line[:-len(self.term)].strip()
                        sqlcmd.append(indent+line)
                        result.append(self.__onecmd(indent, SQLCMD, ' '.join(sqlcmd)))
                    else:
                        sqlcmd.append(line.strip())
                        # It's possible that last line is SQL command without terminator
                        if i+1 == n:
                            result.append(self.__onecmd(indent, SQLCMD, ' '.join(sqlcmd)))
                i = i + 1
        except pp.ParseBaseException, err:
            self.shell.writeErr(err.line)
            msg = str(err.msg)
            pos = msg.find('(at char')
            if pos > -1:
                msg = msg[:pos-1]
            self.shell.writeErr(" "*(err.column-1) + "^")
            self.shell.writeErr(msg)
            return '', False, True
        return '\n'.join(result), multilinesql, False
    
#---Initialization---#000000#FFFFFF---------------------------------------------

